import sys
import argparse
import logging
from rich.logging import RichHandler
import yaml

from gaarf.api_clients import GOOGLE_ADS_API_VERSION
from gaarf.io import writer, reader  # type: ignore
from gaarf.simulation import simulate_data, SimulatorSpecification
from .utils import GaarfConfigBuilder, init_logging


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="+")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("-s",
                        "--simulator-config",
                        dest="simulator_config",
                        default=None)
    parser.add_argument("--account", dest="customer_id", default="None")
    parser.add_argument("--output", dest="save", default="console")
    parser.add_argument("--input", dest="input", default="file")
    parser.add_argument("--api-version", dest="api_version", default=GOOGLE_ADS_API_VERSION)
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("--logger", dest="logger", default="local")
    parser.add_argument("--customer-ids-query",
                        dest="customer_ids_query",
                        default=None)
    parser.add_argument("--customer-ids-query-file",
                        dest="customer_ids_query_file",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.set_defaults(save_config=False)
    args = parser.parse_known_args()
    main_args = args[0]

    logger = init_logging(loglevel=main_args.loglevel.upper(),
                          logger_type=main_args.logger)

    config = GaarfConfigBuilder(args).build()
    logger.debug("config: %s", config)

    writer_client = writer.WriterFactory().create_writer(
        config.output, **config.writer_params)
    if config.output == "bq":
        _ = writer_client.create_or_get_dataset()
    reader_factory = reader.ReaderFactory()
    reader_client = reader_factory.create_reader(main_args.input)

    for query in main_args.query:
        if simulator_config := main_args.simulator_config:
            with open(simulator_config, encoding="utf-8") as f:
                simulator_config = yaml.safe_load(f)
            simulator_specification = SimulatorSpecification(
                **simulator_config)
        else:
            simulator_specification = SimulatorSpecification()
        logger.info("Simulating data for query %s", query)
        report = simulate_data(reader_client.read(query),
                               simulator_specification)
        writer_client.write(report, query)


if __name__ == "__main__":
    main()



import gaarf.io.formatter as formatter  # type: ignore


def test_results_formatter_format_with_empty_results():
    results = formatter.ResultsFormatter.format([])
    assert results == []


def test_results_formatter_format_with_non_empty_results():
    results = formatter.ResultsFormatter.format([1, 2, 3])
    assert results == [[1], [2], [3]]



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from collections.abc import MutableSequence
from concurrent import futures
import functools
import logging
from pathlib import Path
import sys
from rich.logging import RichHandler
from smart_open import open
import yaml

from gaarf import api_clients, query_executor
from gaarf.io import writer, reader  # type: ignore
from .utils import (GaarfConfigBuilder, ConfigSaver,
                    initialize_runtime_parameters, gaarf_runner, init_logging)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="*")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("--account", dest="customer_id", default=None)
    parser.add_argument("--output", dest="save", default="console")
    parser.add_argument("--input", dest="input", default="file")
    parser.add_argument("--ads-config",
                        dest="config",
                        default=str(Path.home() / "google-ads.yaml"))
    parser.add_argument("--api-version",
                        dest="api_version",
                        default=api_clients.GOOGLE_ADS_API_VERSION)
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("--logger", dest="logger", default="local")
    parser.add_argument("--customer-ids-query",
                        dest="customer_ids_query",
                        default=None)
    parser.add_argument("--customer-ids-query-file",
                        dest="customer_ids_query_file",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.add_argument("--parallel-queries",
                        dest="parallel_queries",
                        action="store_true")
    parser.add_argument("--no-parallel-queries",
                        dest="parallel_queries",
                        action="store_false")
    parser.add_argument("--optimize-performance",
                        dest="optimize_performance",
                        default="NONE")
    parser.add_argument("--dry-run", dest="dry_run", action="store_true")
    parser.add_argument("--disable-account-expansion",
                        dest="disable_account_expansion",
                        action="store_true")
    parser.add_argument("-v", "--version", dest="version", action="store_true")
    parser.set_defaults(save_config=False)
    parser.set_defaults(parallel_queries=True)
    parser.set_defaults(dry_run=False)
    parser.set_defaults(disable_account_expansion=False)
    args = parser.parse_known_args()
    main_args = args[0]

    if main_args.version:
        import pkg_resources
        version = pkg_resources.require(
            "google-ads-api-report-fetcher")[0].version
        print(f"gaarf version {version}")
        exit()

    logger = init_logging(loglevel=main_args.loglevel.upper(),
                          logger_type=main_args.logger)
    if not main_args.query:
        logger.error("Please provide one or more queries to run")
        exit()

    with open(main_args.config, "r", encoding="utf-8") as f:
        google_ads_config_dict = yaml.safe_load(f)

    config = GaarfConfigBuilder(args).build()
    if account := main_args.customer_id:
        config.account = account
    if not config.account:
        if mcc := google_ads_config_dict.get("login_customer_id"):
            config.account = str(mcc)
        else:
            raise ValueError(
                "No account found, please specify via --account CLI flag"
                "or add as login_customer_id in google-ads.yaml")
    logger.debug("config: %s", config)

    if main_args.save_config and not main_args.gaarf_config:
        ConfigSaver(main_args.save_config_dest).save(config)
    if main_args.dry_run:
        exit()

    if config.params:
        config = initialize_runtime_parameters(config)
    logger.debug("initialized config: %s", config)

    ads_client = api_clients.GoogleAdsApiClient(
        config_dict=google_ads_config_dict, version=config.api_version)
    ads_query_executor = query_executor.AdsQueryExecutor(ads_client)
    reader_factory = reader.ReaderFactory()
    reader_client = reader_factory.create_reader(main_args.input)

    if config.customer_ids_query:
        customer_ids_query = config.customer_ids_query
    elif config.customer_ids_query_file:
        file_reader = reader_factory.create_reader("file")
        customer_ids_query = file_reader.read(config.customer_ids_query_file)
    else:
        customer_ids_query = None

    if main_args.disable_account_expansion:
        logger.info(
            "Skipping account expansion because of disable_account_expansion flag"
        )
        customer_ids = config.account if isinstance(
            config.account, MutableSequence) else [config.account]
    else:
        customer_ids = ads_query_executor.expand_mcc(config.account,
                                                     customer_ids_query)
    if customer_ids:
        writer_client = writer.WriterFactory().create_writer(
            config.output, **config.writer_params)
        if config.output == "bq":
            _ = writer_client.create_or_get_dataset()
        if config.output == "sheet":
            writer_client.init_client()

        logger.info("Total number of customer_ids is %d, accounts=[%s]",
                    len(customer_ids), ",".join(map(str, customer_ids)))

        if main_args.parallel_queries:
            logger.info("Running queries in parallel")
            with futures.ThreadPoolExecutor() as executor:
                future_to_query = {
                    executor.submit(ads_query_executor.execute,
                                    reader_client.read(query), query, customer_ids, writer_client, config.params, main_args.optimize_performance):
                    query
                    for query in main_args.query
                }
                for future in futures.as_completed(future_to_query):
                    query = future_to_query[future]
                    gaarf_runner(query, future.result, logger)
        else:
            logger.info("Running queries sequentially")
            for query in main_args.query:
                callback = functools.partial(ads_query_executor.execute,
                                             reader_client.read(query), query,
                                             customer_ids, writer_client,
                                             config.params,
                                             main_args.optimize_performance)
                gaarf_runner(query, callback, logger)
    else:
        logger.warning(
            "Not a single under MCC %s is found that satisfies the following customer_id query: '%s'",
            config.account, customer_ids_query)


if __name__ == "__main__":
    main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional
import logging
import sqlalchemy
import pandas as pd

from .query_post_processor import PostProcessorMixin

logger = logging.getLogger(__name__)


class SqlAlchemyQueryExecutor(PostProcessorMixin):

    def __init__(self, engine: sqlalchemy.engine.base.Engine) -> None:
        self.engine = engine

    def execute(
            self,
            script_name: str,
            query_text: str,
            params: Optional[Dict[str, Any]] = None) -> Optional[pd.DataFrame]:
        query_text = self.replace_params_template(query_text, params)
        with self.engine.begin() as conn:
            conn.connection.executescript(query_text)
        return None



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import argparse
from concurrent import futures
import logging
from rich.logging import RichHandler
import sqlalchemy

from gaarf.io import reader  # type: ignore
from .utils import (GaarfBqConfigBuilder, GaarfSqlConfigBuilder,
                    GaarfBqConfigException, GaarfSqlConfigException,
                    ConfigSaver, initialize_runtime_parameters)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", nargs="+")
    parser.add_argument("-c", "--config", dest="gaarf_config", default=None)
    parser.add_argument("--conn",
                        "--connection-string",
                        dest="connection_string",
                        default=None)
    parser.add_argument("--project", dest="project", default=None)
    parser.add_argument("--dataset-location",
                        dest="dataset_location",
                        default=None)
    parser.add_argument("--save-config",
                        dest="save_config",
                        action="store_true")
    parser.add_argument("--no-save-config",
                        dest="save_config",
                        action="store_false")
    parser.add_argument("--config-destination",
                        dest="save_config_dest",
                        default="config.yaml")
    parser.add_argument("--log", "--loglevel", dest="loglevel", default="info")
    parser.add_argument("--logger", dest="logger", default="local")
    parser.add_argument("--dry-run", dest="dry_run", action="store_true")
    parser.set_defaults(save_config=False)
    parser.set_defaults(dry_run=False)
    args = parser.parse_known_args()
    main_args = args[0]

    logger = init_logging(main_args.logger)
    try:
        config = GaarfBqConfigBuilder(args).build()
    except GaarfBqConfigException:
        try:
            config = GaarfSqlConfigBuilder(args).build()
        except GaarfSqlConfigException:
            raise ValueError("failed to build config")

    logger.debug("config: %s", config)
    if main_args.save_config and not main_args.gaarf_config:
        ConfigSaver(main_args.save_config_dest).save(config)
    if main_args.dry_run:
        exit()

    config = initialize_runtime_parameters(config)
    logger.debug("initialized config: %s", config)

    if config.project:
        from gaarf.bq_executor import BigQueryExecutor
        postprocessor = BigQueryExecutor(project_id=config.project,
                                         location=config.dataset_location)
        postprocessor.create_datasets(config.params.get("macro"))
    elif config.connection_string:
        from gaarf.sql_executor import SqlAlchemyQueryExecutor
        engine = sqlalchemy.create_engine(config.connection_string)
        postprocessor = SqlAlchemyQueryExecutor(engine)
    else:
        raise ValueError("Failed to instantiate query postprocessor")

    reader_client = reader.FileReader()

    with futures.ThreadPoolExecutor() as executor:
        future_to_query = {
            executor.submit(postprocessor.execute, query,
                            reader_client.read(query), config.params):
            query
            for query in main_args.query
        }
        for future in futures.as_completed(future_to_query):
            query = future_to_query[future]
            try:
                logger.debug("starting query %s", query)
                future.result()
                logger.info("%s executed successfully", query)
            except Exception as e:
                logger.error("%s generated an exception: %s", query, str(e))


if __name__ == "__main__":
    main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Sequence, Tuple
import logging
import abc
import os
import proto  # type: ignore
import datetime

from google.cloud import bigquery  # type: ignore
from google.cloud.exceptions import NotFound  # type: ignore
from pathlib import Path
import csv
import rich
from rich.console import Console
from rich.table import Table
import pandas as pd  # type: ignore
import numpy as np

from ..report import GaarfReport
from .formatter import ArrayFormatter, ResultsFormatter  # type: ignore

logger = logging.getLogger(__name__)


class AbsWriter(abc.ABC):

    @abc.abstractmethod
    def write(self, results: GaarfReport, destination: str) -> str:
        pass

    def get_columns_results(
            self, results: GaarfReport) -> Tuple[Sequence[str], Sequence[Any]]:
        column_names = results.column_names
        formatted_results = ArrayFormatter.format(
            ResultsFormatter.format(results.to_list()), "|")
        return column_names, formatted_results


class StdoutWriter(AbsWriter):

    def __init__(self, page_size: int = 10, **kwargs):
        self.page_size = int(page_size)

    def write(self, results, destination):
        console = Console()
        table = Table(
            title=f"showing results for query <{destination.split('/')[-1]}>",
            caption=
            f"showing rows 1-{min(self.page_size, len(results))} out of total {len(results)}",
            box=rich.box.MARKDOWN)
        column_names, results = self.get_columns_results(results)
        for header in column_names:
            table.add_column(header)
        for i, row in enumerate(results):
            if i < self.page_size:
                table.add_row(*[str(field) for field in row])
        console.print(table)


class CsvWriter(AbsWriter):

    def __init__(self,
                 destination_folder=os.getcwd(),
                 delimiter=",",
                 quotechar='"',
                 quoting=csv.QUOTE_MINIMAL,
                 **kwargs):
        self.destination_folder = destination_folder
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.quoting = quoting

    def __str__(self):
        return f"[CSV] - data are saved to {self.destination_folder} destination_folder."

    def write(self, results, destination) -> str:
        column_names, results = self.get_columns_results(results)
        destination = DestinationFormatter.format_extension(
            destination, new_extension=".csv")
        if not os.path.isdir(self.destination_folder):
            os.makedirs(self.destination_folder)
        with open(os.path.join(self.destination_folder, destination),
                  "w") as file:
            writer = csv.writer(file,
                                delimiter=self.delimiter,
                                quotechar=self.quotechar,
                                quoting=self.quoting)
            writer.writerow(column_names)
            writer.writerows(results)
        return f"[CSV] - at {destination}"


class SheetWriter(AbsWriter):

    def __init__(self,
                 share_with: str,
                 credentials_file: str,
                 spreadsheet_url: Optional[str] = None,
                 is_append: bool = False,
                 **kwargs: str) -> None:
        """Initialize the SheetWriter to write reports to Google Sheets.

        Args:
            share_with: Email address to share the spreadsheet.
            credentials_file: Path to the service account credentials file.
            spreadsheet_url: URL of the Google Sheets spreadsheet.
            is_append: Whether you want to append data to the spreadsheet.
        """
        self.share_with = share_with
        self.credentials_file = credentials_file
        self.spreadsheet_url = spreadsheet_url
        self.is_append = is_append
        self.client = None
        self.spreadsheet = None

    def __str__(self):
        return f"[Sheet] - data are saved to {self.sheet_url}."

    def init_client(self) -> None:
        import gspread

        if not self.client:
            scope = [
                'https://spreadsheets.google.com/feeds',
                'https://www.googleapis.com/auth/drive'
            ]
            if not self.credentials_file:
                raise ValueError(
                    "Provide path to service account via `credentials_file` option"
                )
            self.gspread_client = gspread.service_account(
                filename=self.credentials_file)
            self._open_sheet()

    def _open_sheet(self) -> None:
        if not self.spreadsheet:
            if not self.spreadsheet_url:
                self.spreadsheet = self.gspread_client.create(
                    f'Gaarf CSV {datetime.datetime.utcnow()}')
            else:
                self.spreadsheet = self.gspread_client.open_by_url(
                    self.spreadsheet_url)
            if self.share_with:
                self.spreadsheet.share(self.share_with,
                                       perm_type='user',
                                       role='writer')

    def write(self,
              results,
              destination=f'Report {datetime.datetime.utcnow()}') -> str:
        import gspread
        self.init_client()
        if not destination:
            destination = f'Report {datetime.datetime.utcnow()}'
        destination = DestinationFormatter.format_extension(destination)
        column_names, results = self.get_columns_results(results)
        num_data_rows = len(results) + 1
        try:
            sheet = self.spreadsheet.worksheet(destination)
        except gspread.exceptions.WorksheetNotFound:
            sheet = self.spreadsheet.add_worksheet(destination,
                                                   rows=num_data_rows,
                                                   cols=len(column_names))
        if not self.is_append:
            sheet.clear()
            self.add_rows_if_needed(num_data_rows, sheet)
            sheet.append_rows([column_names] + results,
                              value_input_option='RAW')
        else:
            self.add_rows_if_needed(num_data_rows, sheet)
            sheet.append_rows(results, value_input_option='RAW')

        success_msg = f"Report is saved to {sheet.url}"
        logger.info(success_msg)
        return success_msg

    def add_rows_if_needed(self, num_data_rows: int,
                           sheet: "gspread.worksheet.Worksheet") -> None:
        num_sheet_rows = len(sheet.get_all_values())
        if num_data_rows > num_sheet_rows:
            num_rows_to_add = num_data_rows - num_sheet_rows
            sheet.add_rows(num_rows_to_add)


class BigQueryWriter(AbsWriter):

    def __init__(self,
                 project: str,
                 dataset: str,
                 location: str = "US",
                 write_disposition: bigquery.WriteDisposition = bigquery.
                 WriteDisposition.WRITE_TRUNCATE,
                 **kwargs):
        self.project = project
        self.dataset_id = f"{project}.{dataset}"
        self.location = location
        self.write_disposition = write_disposition
        self.client = None

    def __str__(self):
        return f"[BigQuery] - {self.dataset_id} at {self.location} location."

    def create_or_get_dataset(self):
        self._init_client()
        try:
            bq_dataset = self.client.get_dataset(self.dataset_id)
        except NotFound:
            bq_dataset = bigquery.Dataset(self.dataset_id)
            bq_dataset.location = self.location
            bq_dataset = self.client.create_dataset(bq_dataset, timeout=30)
        return bq_dataset

    def write(self, results, destination) -> str:
        self._init_client()
        fake_report = results.is_fake
        column_names, results = self.get_columns_results(results)
        schema = self._define_schema(iter(results), column_names)
        destination = DestinationFormatter.format_extension(destination)
        table = self._create_or_get_table(f"{self.dataset_id}.{destination}",
                                          schema)
        job_config = bigquery.LoadJobConfig(
            write_disposition=self.write_disposition,
            schema=schema)

        df = pd.DataFrame(results, columns=column_names)
        df = df.replace({np.nan: None})
        if fake_report:
            df = df.head(0)
        logger.debug("Writing %d rows of data to %s", len(df), destination)
        try:
            self.client.load_table_from_dataframe(dataframe=df,
                                                  destination=table,
                                                  job_config=job_config)
            logger.debug("Writing to %s is completed", destination)
        except Exception as e:
            raise ValueError(f"Unable to save data to BigQuery! {str(e)}")
        return f"[BigQuery] - at {self.dataset_id}.{destination}"

    @staticmethod
    def _define_schema(results, header) -> List[bigquery.SchemaField]:
        result_types = BigQueryWriter._get_result_types(results, header)
        return BigQueryWriter._get_bq_schema(result_types)

    @staticmethod
    def _get_result_types(
            elements: Sequence[Any],
            column_names: Sequence[str]) -> Dict[str, Dict[str, Any]]:
        result_types = {}
        for row in elements:
            if set(column_names) == set(result_types.keys()):
                break
            for i, field in enumerate(row):
                if field is None or column_names[i] in result_types:
                    continue
                field_type = type(field)
                if field_type in [
                        list,
                        proto.marshal.collections.repeated.RepeatedComposite,
                        proto.marshal.collections.repeated.Repeated
                ]:
                    repeated = True
                    if len(field) == 0:
                        field_type = str
                    else:
                        field_type = type(field[0])
                else:
                    field_type = type(field)
                    repeated = False
                result_types[column_names[i]] = {
                    "field_type": field_type,
                    "repeated": repeated
                }
        return result_types

    @staticmethod
    def _get_bq_schema(types) -> List[bigquery.SchemaField]:
        TYPE_MAPPING = {
            list: "REPEATED",
            str: "STRING",
            int: "INT64",
            float: "FLOAT64",
            bool: "BOOL",
            proto.marshal.collections.repeated.RepeatedComposite: "REPEATED",
            proto.marshal.collections.repeated.Repeated: "REPEATED"
        }

        schema: List[bigquery.SchemaField] = []
        for key, value in types.items():
            field_type = TYPE_MAPPING.get(value.get("field_type"))
            schema.append(
                bigquery.SchemaField(
                    name=key,
                    field_type="STRING" if not field_type else field_type,
                    mode="REPEATED" if value.get("repeated") else "NULLABLE"))
        return schema

    def _init_client(self) -> None:
        if not self.client:
            self.client = bigquery.Client(self.project)

    def _create_or_get_table(self, table_name: str, schema) -> bigquery.Table:
        self._init_client()
        try:
            table = self.client.get_table(table_name)
        except NotFound:
            table_ref = bigquery.Table(table_name, schema=schema)
            table = self.client.create_table(table_ref)
            table = self.client.get_table(table_name)
        return table


class SqlAlchemyWriter(AbsWriter):

    def __init__(self,
                 connection_string: str,
                 if_exists: str = "replace",
                 **kwargs):
        self.connection_string = connection_string
        self.if_exists = if_exists

    def write(self, results, destination):
        fake_report = results.is_fake
        column_names, results = self.get_columns_results(results)
        destination = DestinationFormatter.format_extension(destination)
        df = pd.DataFrame(data=results, columns=column_names)
        if fake_report:
            df = df.head(0)
        logger.debug("Writing %d rows of data to %s", len(df), destination)
        df.to_sql(name=destination,
                  con=self._create_engine(),
                  if_exists=self.if_exists)
        logger.debug("Writing to %s is completed", destination)

    def _create_engine(self):
        from sqlalchemy import create_engine
        return create_engine(self.connection_string)


class NullWriter(AbsWriter):

    def __init__(self, writer_option, **kwargs):
        raise ValueError(f"{writer_option} is unknown writer type!")

    def write(self):
        print("Unknown writer type!")


class WriterFactory:
    write_options: Dict[str, AbsWriter] = {}

    def __init__(self):
        self.load_writer_options()

    def load_writer_options(self):
        self.write_options["bq"] = BigQueryWriter
        self.write_options["csv"] = CsvWriter
        self.write_options["sheet"] = SheetWriter
        self.write_options["console"] = StdoutWriter
        self.write_options["sqldb"] = SqlAlchemyWriter

    def create_writer(self, writer_option, **kwargs):
        if writer_option in self.write_options:
            return self.write_options[writer_option](**kwargs)
        else:
            return NullWriter(writer_option)


class DestinationFormatter:

    @staticmethod
    def format_extension(path_object: str,
                         current_extension: str = ".sql",
                         new_extension: str = "") -> str:
        path_object_name = Path(path_object).name
        if len(path_object_name.split(".")) > 1:
            return path_object_name.replace(current_extension, new_extension)
        else:
            return f"{path_object}{new_extension}"


class ZeroRowException(Exception):
    pass



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import proto

from typing import Callable, Dict
import re

from .report import GaarfRow, GaarfReport


def get_ocid_mapping(report_fetcher, accounts) -> GaarfReport:
    query = "SELECT customer.id, metrics.optimization_score_url FROM campaign LIMIT 1"
    mapping = []
    for account in accounts:
        report = report_fetcher.fetch(query, account)
        if report and (ocid := re.findall("ocid=(\w+)", report[0][1])):
            mapping.append((report[0][0], ocid[0]))
        else:
            mapping.append((int(account), "0"))
    return GaarfReport(results=mapping, column_names=("account_id", "ocid"))


BUILTIN_QUERIES: Dict[str, Callable] = {
    "ocid_mapping": get_ocid_mapping
}



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from collections.abc import MutableSequence
from concurrent import futures
from enum import Enum, auto
import itertools
from google.ads.googleads.errors import GoogleAdsException  # type: ignore
import logging
from google.api_core import exceptions
from tenacity import Retrying, RetryError, retry_if_exception_type, stop_after_attempt, wait_exponential
import warnings

from . import parsers
from . import api_clients
from . import builtin_queries
from .query_editor import QuerySpecification, QueryElements
from .report import GaarfReport, GaarfRow
from .io import writer, reader  # type: ignore

logger = logging.getLogger(__name__)


class OptimizeStrategy(Enum):
    NONE = auto()
    BATCH = auto()
    PROTOBUF = auto()
    BATCH_PROTOBUF = auto()


class AdsReportFetcher:
    """Class responsible for getting data from Ads API.

    Attributes:
        api_client: a client used for connecting to Ads API.
    """

    def __init__(self,
                 api_client: api_clients.BaseClient,
                 customer_ids: Optional[Any] = None) -> None:
        self.api_client = api_client
        if customer_ids:
            warnings.warn(
                "`AdsReportFetcher` will deprecate passing `customer_ids` to `__init__` method. "
                "Consider passing list of customer_ids to `AdsReportFetcher.fetch` method",
                category=DeprecationWarning,
                stacklevel=3)
            self.customer_ids = [
                customer_ids
            ] if not isinstance(customer_ids, list) else customer_ids

    def expand_mcc(self,
                   customer_ids: Union[str, MutableSequence[str]],
                   customer_ids_query: Optional[str] = None) -> List[str]:
        return self._get_customer_ids(customer_ids, customer_ids_query)

    def fetch(self,
              query_specification: Union[str, QueryElements],
              customer_ids: Optional[Union[List[str], str]] = None,
              customer_ids_query: Optional[str] = None,
              expand_mcc: bool = False,
              args: Optional[Dict[str, Any]] = None,
              optimize_strategy: str = "NONE") -> GaarfReport:
        """Fetches data from Ads API based on query_specification.

        Args:
            query_specification: Query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_ids: Account(s) for from which data should be fetched.
            custom_query: GAQL query used to reduce the number of customer_ids.
            expand_mcc: Whether to perform expansion of root customer_ids
                into leaf accounts.
            args: Arguments that need to be passed to the query
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF").

        Returns:
            GaarfReport with results of query execution.
        """

        if isinstance(self.api_client, api_clients.GoogleAdsApiClient):
            if not customer_ids:
                warnings.warn(
                    "`AdsReportFetcher` will require passing `customer_ids` to `fetch` method.",
                    category=DeprecationWarning,
                    stacklevel=3)
                if hasattr(self, "customer_ids"):
                    if not self.customer_ids:
                        raise ValueError(
                            "Please specify add `customer_ids` to `fetch` method"
                        )
                    customer_ids = self.customer_ids
            else:
                if expand_mcc:
                    customer_ids = self.expand_mcc(customer_ids,
                                                   customer_ids_query)
                customer_ids = [
                    customer_ids
                ] if not isinstance(customer_ids, list) else customer_ids
        else:
            customer_ids = []
        total_results: List[List[Tuple[Any]]] = []
        is_fake_report = False
        if not isinstance(query_specification, QueryElements):
            query_specification = QuerySpecification(
                text=str(query_specification),
                args=args,
                api_version=self.api_client.api_version).generate()

        if query_specification.is_builtin_query:
            if not (report := builtin_queries.BUILTIN_QUERIES.get(
                    query_specification.query_title)):
                raise ValueError(
                    f"Cannot find the built-in query '{query_specification.title}'"
                )
            return report(self, accounts=customer_ids)
        parser = parsers.GoogleAdsRowParser(query_specification)
        for customer_id in customer_ids:
            logger.debug("Running query %s for customer_id %s",
                         query_specification.query_title, customer_id)
            try:
                results = self._parse_ads_response(
                    query_specification, customer_id, parser,
                    getattr(OptimizeStrategy, optimize_strategy))
                total_results.extend(results)
                if query_specification.is_constant_resource:
                    logger.debug("Constant resource query: running only once")
                    break
            except GoogleAdsException as e:
                logger.error("Cannot execute query %s for %s",
                             query_specification.query_title, customer_id)
                logger.error(str(e))
                raise
        if not total_results:
            row = self.api_client.google_ads_row
            results = [parser.parse_ads_row(row)]
            total_results.extend(results)
            is_fake_report = True
            if not isinstance(self.api_client, api_clients.BaseClient):
                logger.warning(
                    "Query %s generated zero results, using placeholders to infer schema",
                    query_specification.query_title)
        return GaarfReport(results=total_results,
                           column_names=query_specification.column_names,
                           is_fake=is_fake_report,
                           query_specification=query_specification)

    def _parse_ads_response(
        self,
        query_specification: QueryElements,
        customer_id: str,
        parser: parsers.GoogleAdsRowParser,
        optimize_strategy: OptimizeStrategy = OptimizeStrategy.NONE
    ) -> List[List[Tuple[Any]]]:
        """Parses response returned from Ads API request.

        Args:
            query_specification: query text that will be passed to Ads API
                alongside column_names, customizers and virtual columns.
            customer_id: account for which data should be requested.
            parser: instance of parser class that transforms each row from
                request into desired format.
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF").

        Returns:
            Parsed rows for the whole response.
        """
        total_results: List[List[Tuple[Any]]] = []
        logger.debug("Getting response for query %s for customer_id %s",
                     query_specification.query_title, customer_id)
        try:
            for attempt in Retrying(retry=retry_if_exception_type(
                    exceptions.InternalServerError),
                                    stop=stop_after_attempt(3),
                                    wait=wait_exponential()):
                with attempt:
                    response = self.api_client.get_response(
                        entity_id=str(customer_id),
                        query_text=query_specification.query_text)
        except RetryError as retry_failure:
            logger.error("Cannot fetch data from API for query '%s' %d times",
                         query_specification.query_title,
                         retry_failure.last_attempt.attempt_number)
            raise exceptions.InternalServerError
        if optimize_strategy in (OptimizeStrategy.BATCH,
                                 OptimizeStrategy.BATCH_PROTOBUF):
            logger.warning("Running gaarf in an optimized mode")
            logger.warning("Optimize strategy is %s", optimize_strategy.name)
            if optimize_strategy == OptimizeStrategy.BATCH_PROTOBUF:
                rows = [row._pb for batch in response for row in batch.results]
            else:
                rows = [row for batch in response for row in batch.results]
            parsed_batches = []
            with futures.ThreadPoolExecutor() as executor:
                BATCH_SIZE = 10000
                batches = list(range(0, len(rows), BATCH_SIZE))
                future_to_batch = {}
                for i, batch_index in enumerate(batches):
                    try:
                        from_, to = batch_index, batches[i + 1]
                        future_to_batch[executor.submit(
                            self._parse_batch, parser,
                            itertools.islice(rows, from_,
                                             to))] = itertools.islice(
                                                 rows, from_, to)
                    except IndexError:
                        future_to_batch[executor.submit(
                            self._parse_batch, parser,
                            itertools.islice(rows, batch_index,
                                             len(rows)))] = itertools.islice(
                                                 rows, batch_index, len(rows))

                for i, future in enumerate(
                        futures.as_completed(future_to_batch), start=1):
                    batch = future_to_batch[future]
                    logger.debug(
                        "Parsed batch %d for query %s for customer_id %s", i,
                        query_specification.query_title, customer_id)
                    parsed_batch = future.result()
                    parsed_batches.append(parsed_batch)
            total_results = [
                items for batch in parsed_batches for items in batch
            ]

        else:
            logger.debug(
                "Iterating over response for query %s for customer_id %s",
                query_specification.query_title, customer_id)
            for batch in response:
                logger.debug("Parsing batch for query %s for customer_id %s",
                             query_specification.query_title, customer_id)
                if optimize_strategy == OptimizeStrategy.PROTOBUF:
                    logger.warning("Running gaarf in an optimized mode")
                    logger.warning("Optimize strategy is %s",
                                   optimize_strategy.name)
                    rows = (row._pb for row in batch.results)
                else:
                    rows = (row for row in batch.results)
                results = self._parse_batch(parser, rows)
                total_results.extend(results)
        return total_results

    def _parse_batch(
            self, parser,
            batch: Sequence["GoogleAdsRow"]) -> List[List[Tuple[Any]]]:
        """Parse reach row from batch of Ads API response.

        Args:
            parser: instance of parser class that transforms each row from
                request into desired format.
            batch: sequence of GoogleAdsRow that needs to be parsed.
        Returns:
            Parsed rows for a batch.
        """

        return [parser.parse_ads_row(row) for row in batch]

    def _get_customer_ids(
            self,
            seed_customer_ids: Union[str, MutableSequence[str]],
            customer_ids_query: Optional[str] = None) -> List[str]:
        """Gets list of customer_ids from an MCC account.

        Args:
            customer_ids: MCC account_id(s).
            custom_query: GAQL query used to reduce the number of customer_ids.
        Returns:
            All customer_ids from MCC satisfying the condition.
        """

        query = """
        SELECT customer_client.id FROM customer_client
        WHERE customer_client.manager = FALSE AND customer_client.status = "ENABLED"
        """
        query_specification = QuerySpecification(query).generate()
        if not isinstance(seed_customer_ids, MutableSequence):
            seed_customer_ids = seed_customer_ids.split(",")
        child_customer_ids = self.fetch(query_specification,
                                        seed_customer_ids).to_list()
        if customer_ids_query:
            query_specification = QuerySpecification(
                customer_ids_query).generate()
            child_customer_ids = self.fetch(query_specification,
                                            child_customer_ids)
            child_customer_ids = [
                row[0] if isinstance(row, GaarfRow) else row
                for row in child_customer_ids
            ]

        child_customer_ids = list(
            set([
                customer_id for customer_id in child_customer_ids
                if customer_id != 0
            ]))

        return child_customer_ids


class AdsQueryExecutor:
    """Class responsible for getting data from Ads API and writing it to
        local/remote storage.

    Attributes:
        api_client: a client used for connecting to Ads API.
    """

    def __init__(self, api_client: api_clients.BaseClient):
        self.report_fetcher = AdsReportFetcher(api_client)

    def execute(self,
                query_text: str,
                query_name: str,
                customer_ids: Union[List[str], str],
                writer_client: writer.AbsWriter = writer.StdoutWriter(),
                args: Optional[Dict[str, Any]] = None,
                optimize_performance: str = "NONE") -> None:
        """Reads query, extract results and stores them in a specified location.

        Args:
            query_text: Text for the query.
            customer_ids: All accounts for which query will be executed.
            writer_client: Client responsible for writing data to local/remote
                location.
            args: Arguments that need to be passed to the query
            optimize_strategy: strategy for speeding up query execution
                ("NONE", "PROTOBUF", "BATCH", "BATCH_PROTOBUF")
        """

        query_specification = QuerySpecification(
            query_text, query_name, args,
            self.report_fetcher.api_client.api_version).generate()
        results = self.report_fetcher.fetch(
            query_specification=query_specification,
            customer_ids=customer_ids,
            optimize_strategy=optimize_performance)
        logger.debug("Start writing data for query %s via %s writer",
                     query_specification.query_title, type(writer_client))
        writer_client.write(results, query_specification.query_title)
        logger.debug("Finish writing data for query %s via %s writer",
                     query_specification.query_title, type(writer_client))

    def expand_mcc(self,
                   customer_ids: Union[str, MutableSequence[str]],
                   customer_ids_query: Optional[str] = None) -> List[str]:
        return self.report_fetcher._get_customer_ids(customer_ids,
                                                     customer_ids_query)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Sequence, Union
from collections.abc import MutableSequence
from dataclasses import dataclass, field, asdict
import logging
from rich.logging import RichHandler
import os
import sys
import datetime
from smart_open import open
from dateutil.relativedelta import relativedelta
import yaml
from google.ads.googleads.errors import GoogleAdsException  # type: ignore
import traceback

from gaarf.io.writer import ZeroRowException
from gaarf.query_editor import CommonParametersMixin


@dataclass
class GaarfConfig:
    output: str
    api_version: str
    account: str
    params: Dict[str, Any] = field(default_factory=dict)
    writer_params: Dict[str, Any] = field(default_factory=dict)
    customer_ids_query: Optional[str] = None
    customer_ids_query_file: Optional[str] = None


class GaarfConfigException(Exception):
    ...


@dataclass
class GaarfBqConfig:
    project: str
    dataset_location: Optional[str]
    params: Dict[str, Any] = field(default_factory=dict)


class GaarfBqConfigException(Exception):
    ...


@dataclass
class GaarfSqlConfig:
    connection_string: str
    params: Dict[str, Any] = field(default_factory=dict)


class GaarfSqlConfigException(Exception):
    ...


class BaseConfigBuilder:

    def __init__(self, args):
        self.args = args
        if (gaarf_config_path := self.args[0].gaarf_config):
            self.type = "file"
            self.gaarf_config_path = gaarf_config_path
        else:
            self.type = "cli"
            self.gaarf_config_path = None

    def build(self) -> Union[GaarfConfig, GaarfBqConfig]:
        if self.type == "file":
            return self._load_gaarf_config()
        return self._build_gaarf_config()

    def _load_gaarf_config(self) -> Union[GaarfConfig, GaarfBqConfig]:
        raise NotImplementedError

    def _build_gaarf_config(self) -> Union[GaarfConfig, GaarfBqConfig]:
        raise NotImplementedError


class GaarfConfigBuilder(BaseConfigBuilder):

    def __init__(self, args):
        super().__init__(args)

    def _load_gaarf_config(self) -> GaarfConfig:
        with open(self.gaarf_config_path, encoding="utf-8") as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get("gaarf")
        if not gaarf_section:
            raise GaarfConfigException(
                "Invalid config, must have `gaarf` section!")
        if not (output := gaarf_section.get("output")):
            raise ValueError("Config does not contains `output` section!")
        return GaarfConfig(
            output=gaarf_section.get("output"),
            api_version=gaarf_section.get("api_version"),
            account=gaarf_section.get("account"),
            params=gaarf_section.get("params"),
            writer_params=gaarf_section.get(output),
            customer_ids_query=gaarf_section.get("customer_ids_query"),
            customer_ids_query_file=gaarf_section.get(
                "customer_ids_query_file"))

    def _build_gaarf_config(self) -> GaarfConfig:
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser(["macro", "template",
                               main_args.save]).parse(query_args)
        return GaarfConfig(
            output=main_args.save,
            api_version=main_args.api_version,
            account=main_args.customer_id,
            params=params,
            writer_params=params.get(main_args.save),
            customer_ids_query=main_args.customer_ids_query,
            customer_ids_query_file=main_args.customer_ids_query_file)


class GaarfBqConfigBuilder(BaseConfigBuilder):

    def __init__(self, args):
        super().__init__(args)

    def _load_gaarf_config(self) -> GaarfBqConfig:
        with open(self.gaarf_config_path, encoding="utf-8") as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get("gaarf-bq")
        if not gaarf_section:
            raise GaarfBqConfigException(
                "Invalid config, must have `gaarf-bq` section!")
        params = gaarf_section.get("params")
        return GaarfBqConfig(
            project=gaarf_section.get("project"),
            dataset_location=gaarf_section.get("dataset_location"),
            params=params)

    def _build_gaarf_config(self) -> GaarfBqConfig:
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser(["macro", "sql", "template"]).parse(query_args)
        return GaarfBqConfig(project=main_args.project,
                             dataset_location=main_args.dataset_location,
                             params=params)


class GaarfSqlConfigBuilder(BaseConfigBuilder):

    def __init__(self, args):
        import sqlalchemy
        super().__init__(args)

    def _load_gaarf_config(self) -> GaarfBqConfig:
        with open(self.gaarf_config_path, encoding="utf-8") as f:
            config = yaml.safe_load(f)
        gaarf_section = config.get("gaarf-sql")
        if not gaarf_section:
            raise GaarfSqlConfigException(
                "Invalid config, must have `gaarf-sql` section!")
        params = gaarf_section.get("params")
        return GaarfSqlConfig(
            connection_string=gaarf_section.get("connection-string").format(
                **dict(os.environ.items())),
            params=params)

    def _build_gaarf_config(self) -> GaarfSqlConfig:
        main_args, query_args = self.args[0], self.args[1]
        params = ParamsParser(["macro", "sql", "template"]).parse(query_args)
        return GaarfSqlConfig(
            connection_string=main_args.connection_string.format(
                **dict(os.environ.items())),
            params=params)


class ParamsParser(CommonParametersMixin):

    def __init__(self, identifiers: Sequence[str]):
        self.identifiers = identifiers

    def parse(self,
              params: Sequence[Any]) -> Dict[str, Optional[Dict[str, Any]]]:
        return {
            identifier: self._parse_params(identifier, params)
            for identifier in self.identifiers
        }

    def _parse_params(self, identifier: str,
                      params: Sequence[Any]) -> Dict[Any, Any]:
        parsed_params = {}
        if params:
            raw_params = [param.split("=", maxsplit=1) for param in params]
            for param in raw_params:
                param_pair = self._identify_param_pair(identifier, param)
                if param_pair:
                    parsed_params.update(param_pair)
        return parsed_params

    def _identify_param_pair(self, identifier: str,
                             param: Sequence[str]) -> Optional[Dict[str, Any]]:
        key = param[0]
        if identifier not in key:
            return None
        provided_identifier, key = key.split(".")
        if provided_identifier.replace("--", "") not in self.identifiers:
            raise GaarfParamsException(
                f"CLI argument {provided_identifier} is not supported"
                f", supported arguments {', '.join(self.identifiers)}")
        key = key.replace("-", "_")
        if len(param) == 2:
            return {key: param[1]}
        raise GaarfParamsException(
            f"{identifier} {key} is invalid,"
            f"--{identifier}.key=value is the correct format")


class GaarfParamsException(Exception):
    ...


def convert_date(date_string: str) -> str:
    """Converts specific dates parameters to actual dates."""

    if isinstance(date_string, list) or date_string.find(":YYYY") == -1:
        return date_string
    current_date = datetime.date.today()
    date_object = date_string.split("-")
    base_date = date_object[0]
    if len(date_object) == 2:
        try:
            days_ago = int(date_object[1])
        except ValueError as e:
            raise ValueError(
                "Must provide numeric value for a number lookback period, "
                "i.e. :YYYYMMDD-1") from e
    else:
        days_ago = 0
    if base_date == ":YYYY":
        new_date = datetime.datetime(current_date.year, 1, 1)
        delta = relativedelta(years=days_ago)
    elif base_date == ":YYYYMM":
        new_date = datetime.datetime(current_date.year, current_date.month, 1)
        delta = relativedelta(months=days_ago)
    elif base_date == ":YYYYMMDD":
        new_date = current_date
        delta = relativedelta(days=days_ago)
    return (new_date - delta).strftime("%Y-%m-%d")


class ConfigSaver:

    def __init__(self, path: str):
        self.path = path

    def save(self, gaarf_config: Union[GaarfConfig, GaarfBqConfig]):
        if os.path.exists(self.path):
            with open(self.path, "r", encoding="utf-8") as f:
                config = yaml.safe_load(f)
        else:
            config = {}
        config = self.prepare_config(config, gaarf_config)
        with open(self.path, "w", encoding="utf-8") as f:
            yaml.dump(config,
                      f,
                      default_flow_style=False,
                      sort_keys=False,
                      encoding="utf-8")

    def prepare_config(
        self, config: Dict[Any, Any],
        gaarf_config: Union[GaarfConfig, GaarfBqConfig, GaarfSqlConfig]
    ) -> Dict[str, Any]:
        gaarf = asdict(gaarf_config)
        if isinstance(gaarf_config, GaarfConfig):
            gaarf[gaarf_config.output] = gaarf_config.writer_params
            if not isinstance(gaarf_config.account, MutableSequence):
                gaarf["account"] = gaarf_config.account.split(",")
            del gaarf["writer_params"]
            if gaarf_config.writer_params:
                del gaarf["params"][gaarf_config.output]
            gaarf = _remove_empty_values(gaarf)
            config.update({"gaarf": gaarf})
        if isinstance(gaarf_config, GaarfBqConfig):
            gaarf = _remove_empty_values(gaarf)
            config.update({"gaarf-bq": gaarf})
        if isinstance(gaarf_config, GaarfSqlConfig):
            gaarf = _remove_empty_values(gaarf)
            config.update({"gaarf-sql": gaarf})
        return config


def initialize_runtime_parameters(config: Union[GaarfConfig, GaarfBqConfig]):
    for key, param in config.params.items():
        for key_param, value_param in param.items():
            config.params[key][key_param] = convert_date(value_param)
        config.params[key].update(ParamsParser.common_params)
    return config


def _remove_empty_values(dict_object: Dict[str, Any]) -> Dict[str, Any]:
    """Remove all empty elements: strings, dictionaries from a dictionary."""
    if isinstance(dict_object, dict):
        return {
            key: value
            for key, value in ((key, _remove_empty_values(value))
                               for key, value in dict_object.items()) if value
        }
    if isinstance(dict_object, (int, str, MutableSequence)):
        return dict_object


def gaarf_runner(query: str, callback: Callable, logger) -> None:
    try:
        logger.debug("starting query %s", query)
        callback()
        logger.info("%s executed successfully", query)
    except ZeroRowException:
        logger.warning("%s returns 0 rows", query)
    except GoogleAdsException as ex:
        logger.error(
            "%s failed with status %s and includes the following errors:",
            query,
            ex.error.code().name)
        for error in ex.failure.errors:
            logger.error("\tError with message %s .", error.message)
            if error.location:
                for field in error.location.field_path_elements:
                    logger.error("\t\tOn field %s", field.field_name)
    except Exception as e:
        traceback.print_tb(e.__traceback__)
        logger.error("%s generated an exception: %s", query, str(e))


def postprocessor_runner(query: str, callback: Callable, logger) -> None:
    try:
        logger.debug("starting query %s", query)
        callback()
        logger.info("%s executed successfully", query)
    except Exception as e:
        logger.error("%s generated an exception: %s", query, str(e))


def init_logging(loglevel: str = "INFO",
                 logger_type: str = "local",
                 name: str = __name__) -> logging.Logger:
    if logger_type == "rich":
        logging.basicConfig(format="%(message)s",
                            level=loglevel,
                            datefmt="%Y-%m-%d %H:%M:%S",
                            handlers=[RichHandler(rich_tracebacks=True)])
    else:
        logging.basicConfig(
            format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s",
            stream=sys.stdout,
            level=loglevel,
            datefmt="%Y-%m-%d %H:%M:%S")
    logging.getLogger("google.ads.googleads.client").setLevel(logging.WARNING)
    logging.getLogger("smart_open.smart_open_lib").setLevel(logging.WARNING)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)
    return logging.getLogger(name)



import pytest
import datetime
import gaarf.query_editor as query_editor


@pytest.fixture
def query():
    query = """
-- Comment
# Comment
// Comment

SELECT
    1 AS constant,
    '2023-01-01' AS date,
    '{current_date}' AS current_date,
    metrics.clicks / metrics.impressions AS ctr,
    customer.id, --customer_id
    campaign.bidding_strategy_type AS campaign_type, campaign.id:nested AS campaign,
    ad_group.id~1 AS ad_group,
    ad_group_ad.ad.id->asset AS ad,
    metrics.cost_micros * 1e6 AS cost,
    {% if selective == "true" %}
        campaign.selective_optimization AS selective_optimization,
    {% endif %}
from ad_group_ad
"""
    return query


@pytest.fixture
def builtin_query():
    return "SELECT * FROM builtin.ocid_mapping"


def test_builtin_query_returns_valid_specification(builtin_query):
    spec = query_editor.QuerySpecification(title="/tmp/ocid_mapping.sql",
                                           text=builtin_query,
                                           args=None).generate()
    assert spec.is_builtin_query
    assert spec.query_title == "ocid_mapping"
    assert spec.resource_name == "builtin.ocid_mapping"
    assert not spec.column_names


@pytest.fixture
def query_specification(query):
    return query_editor.QuerySpecification(title="sample_query",
                                           text=query,
                                           args=None)


@pytest.fixture
def query_specification_template(query):
    return query_editor.QuerySpecification(
        title="templated_query",
        text=query,
        args={"template": {
            "selective": "true"
        }})


@pytest.fixture
def sample_query(query_specification):
    return query_specification.generate()


@pytest.fixture
def templated_query(query_specification_template):
    return query_specification_template.generate()


class TestRegularQuery:

    def test_correct_title(self, sample_query):
        assert sample_query.query_title == "sample_query"

    def test_extract_correct_fields(self, sample_query):
        assert sample_query.fields == [
            "customer.id", "campaign.bidding_strategy_type", "campaign.id",
            "ad_group.id", "ad_group_ad.ad.id"
        ]

    def test_extract_correct_aliases(self, sample_query):
        assert sample_query.column_names == [
            "constant", "date", "current_date", "ctr", "customer_id",
            "campaign_type", "campaign", "ad_group", "ad", "cost"
        ]

    def test_extract_correct_text(self, sample_query):
        assert sample_query.query_text.lower(
        ) == "select customer.id, campaign.bidding_strategy_type, campaign.id, ad_group.id, ad_group_ad.ad.id, metrics.clicks, metrics.impressions, metrics.cost_micros from ad_group_ad"

    def test_extract_custom_callers(self, sample_query):
        assert sample_query.customizers == {
            "campaign": {
                "type": "nested_field",
                "value": "nested"
            },
            "ad_group": {
                "type": "resource_index",
                "value": 1
            },
            "ad": {
                "type": "pointer",
                "value": "asset"
            }
        }

    def test_extract_query_lines(self, query_specification, sample_query):
        extracted_lines = query_specification.extract_query_lines(
            sample_query.query_text)
        assert extracted_lines == [
            "customer.id", "campaign.bidding_strategy_type", "campaign.id",
            "ad_group.id", "ad_group_ad.ad.id", "metrics.clicks",
            "metrics.impressions", "metrics.cost_micros"
        ]

    def test_extract_correct_resource(self, sample_query):
        assert sample_query.resource_name == "ad_group_ad"

    def test_is_constant_resource(self, sample_query):
        assert sample_query.is_constant_resource == False

    def test_has_virtual_columns(self, sample_query):
        assert sample_query.virtual_columns == {
            "constant":
            query_editor.VirtualColumn(type="built-in", value=1),
            "date":
            query_editor.VirtualColumn(type="built-in", value="2023-01-01"),
            "current_date":
            query_editor.VirtualColumn(
                type="built-in",
                value=datetime.date.today().strftime("%Y-%m-%d")),
            "ctr":
            query_editor.VirtualColumn(
                type="expression",
                value="metrics.clicks / metrics.impressions",
                fields=["metrics.clicks", "metrics.impressions"],
                substitute_expression="{metrics_clicks} / {metrics_impressions}"
            ),
            "cost":
            query_editor.VirtualColumn(
                type="expression",
                value="metrics.cost_micros * 1e6",
                fields=["metrics.cost_micros"],
                substitute_expression="{metrics_cost_micros} * 1e6")
        }

    def test_incorrect_resource_raises_value_error(self):
        query = "SELECT metrics.clicks FROM ad_groups"
        spec = query_editor.QuerySpecification(title="sample_query",
                                               text=query,
                                               args=None)
        with pytest.raises(ValueError):
            spec.generate()

    def test_incorrect_specification_raises_macro_error(self):
        query = "SELECT '${custom_field}', ad_group.id FROM ad_group"
        spec = query_editor.QuerySpecification(title="sample_query",
                                               text=query,
                                               args=None)
        with pytest.raises(query_editor.MacroError):
            spec.generate()

    def test_incorrect_specification_raises_virtual_column_error(self):
        query = "SELECT 1, ad_group.id AS ad_group_id FROM ad_group"
        spec = query_editor.QuerySpecification(title="sample_query",
                                               text=query,
                                               args=None)
        with pytest.raises(query_editor.VirtualColumnError):
            spec.generate()

    def test_incorrect_field_raises_value_error(self):
        query = "SELECT metric.impressions, ad_group.id FROM ad_group"
        spec = query_editor.QuerySpecification(title="sample_query",
                                               text=query,
                                               args=None)
        with pytest.raises(query_editor.FieldError):
            spec.generate()


class TestTemplatedQuery:

    def test_extract_correct_fields(self, templated_query):
        assert templated_query.fields == [
            "customer.id", "campaign.bidding_strategy_type", "campaign.id",
            "ad_group.id", "ad_group_ad.ad.id",
            "campaign.selective_optimization"
        ]

    def test_extract_correct_aliases(self, templated_query):
        assert templated_query.column_names == [
            "constant", "date", "current_date", "ctr", "customer_id",
            "campaign_type", "campaign", "ad_group", "ad", "cost",
            "selective_optimization"
        ]

    def test_extract_correct_text(self, templated_query):
        assert templated_query.query_text.lower(
        ) == "select customer.id, campaign.bidding_strategy_type, campaign.id, ad_group.id, ad_group_ad.ad.id, campaign.selective_optimization, metrics.clicks, metrics.impressions, metrics.cost_micros from ad_group_ad"

    def test_extract_query_lines(self, query_specification, templated_query):
        extracted_lines = query_specification.extract_query_lines(
            templated_query.query_text)
        assert extracted_lines == [
            "customer.id", "campaign.bidding_strategy_type", "campaign.id",
            "ad_group.id", "ad_group_ad.ad.id",
            "campaign.selective_optimization", "metrics.clicks",
            "metrics.impressions", "metrics.cost_micros"
        ]
